#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "atlas_misc.h"
#include "atlas_r1parse.h"
#include "atlas_r2testtime.h"

/*
 * NREP & NREPCT currently unused: I need to modify atlas-parse to allow
 * repetiton setting to be varied by call.  If I make this change, why
 * not modify FORCETIME so that it will delete any result file if
 * FORCETIME is negative.
 * VERB is also currently unused, may fix later
 */
static int FLUSHKB=(-1), L2KB=0, NREPCT=3, NREP=3, VERB=0;

ATL_r1node_t *TimeAllKernels
(
   int L1CacheElts,             /* size of L1 cache in elements */
   int imf,                     /* index into bp->mflop  */
   char pre,                    /* type/precision prefix */
   int  M, int N, int lda,      /* problem dimensions */
   int Fflops,                  /* what to set Force flops to (see below) */
   int FTIME,                   /* 0: use prior output files */
   int flushKB,                 /* 0: no flushing, else mem to flush */
   ATL_r1node_t *kq             /* queue of kernels */
)
/*
 * This routine times all kernels in kq, putting results in kq->mflop[imf]
 * Fflops: for very large problems, set to 0, else the # of flops to force
 * RETURNS: ptr to fastest kernel
 */
{
   ATL_r1node_t *kp, *bestp=NULL;
   int percL1;
   double mf, mfmax=0.0;

   for (kp=kq; kp; kp = kp->next)
   {
      mf = ((double)kp->CacheElts) / ((double)L1CacheElts)*100.0;
      percL1 = mf;
      mf = TimeR2Kernel(1, FTIME, kp, pre, M, N, lda, percL1, Fflops,flushKB);
      kp->mflop[imf] = mf;
      if (mf > mfmax)
      {
         bestp = kp;
         mfmax = mf;
      }
   }
   return(bestp);
}

ATL_r1node_t *TimeAllKernelsForContext
(
   int L1CacheElts,     /* size of L1 cache in elements */
   int imf,             /* index into mflop & see below */
   char pre,            /* type/precision prefix */
   ATL_r1node_t *bp     /* queue of kernels */
)
/*
 * This routine naively assumes that 4*L1CacheSize is a decent marker for
 * the L2 size to use in initial L2-timings.
 *
 * imf: parameter describing type of timings to perform:
 *    0: large out-of-cache, blocked for 85% of L1CacheSize
 *    1: large out-of-cache, blocked for MIN(128K,4*L1CacheSize) (L2-blocked)
 *    2: large out-of-cache, no blocking
 *    3: in-L2 problem, no blocking
 *    4: in-L1 problem, no blocking
 * RETURNS: best-performing kernel in context
 */
{
   ATL_r1node_t *r1p, *r1max=NULL;
   double mf, mfmax=0.0;
   int M, N, lda, percL1, cflush=(FLUSHKB);

   if (imf == 0 || imf == 1 || imf == 2)
   {
      M = (pre == 's') ? 2000 : 1000;
      N = (pre == 'z') ? 500 : 1000;
      cflush = FLUSHKB;
      percL1 = imf ? 400 : 85;
      percL1 = (imf == 2) ? 0 : percL1;
   }
   else   /* Time in-cache data with no blocking */
      cflush = percL1 = 0;
   for (r1p=bp; r1p; r1p = r1p->next)
   {
      if (imf == 3)  /* L2-contained data */
      {
         N = ((16+r1p->YU-1)/r1p->YU)*r1p->YU;
         M = 128*1024/pre2size(pre);
         if (M > 4*L1CacheElts)
            M = 4*L1CacheElts;
         M /= N;
      }
      else if (imf == 4)  /* L1-contained data */
      {
         N = ((8+r1p->YU-1)/r1p->YU)*r1p->YU;
         M = (85*L1CacheElts)/(N*100);
      }
      lda = M+8;
      mf = TimeR2Kernel(0, 0, r1p, pre, M, N, lda, percL1, -1, cflush);
      if (mf > mfmax)
      {
         mfmax = mf;
         r1max = r1p;
      }
      r1p->mflop[imf] = mf;
   }
   return(r1max);
}

static ATL_r1node_t *DelBadTestKernels(char pre, ATL_r1node_t *bp)
/*
 * Deletes all kernels that can't pass basic usage test
 */
{
   int die;
   ATL_r1node_t *p, *prev;
   int m, n, lda, i, j;
   fprintf(stdout, "\nBEGIN BASIC KERNEL TESTS:\n");

   prev = p = bp;
   while(p)
   {
      m = n = lda = 1000;
      if (FLAG_IS_SET(p->flag, R1F_FYU))
      {
         i = p->YU;
         n = ((n+i-1)/i)*i;
      }
      if (p->ldamul)
      {
         j = pre2size(pre);
         i = p->ldamul / j;
         assert(p->ldamul == i*j);
         lda = ((lda+i-1)/i)*i;
      }
      if (R2KernelFailsTest(0, pre, m, n, lda, p))
      {
         fprintf(stdout, "   NUKING bad kernel %s(%d)\n", p->rout, p->ID);
         if (p == bp)
            bp = p = KillR1Node(p);
         else
            prev->next = p = KillR1Node(p);
      }
      else
      {
         fprintf(stdout, "   Kernel %s(%d) passes basic test\n",
                 p->rout, p->ID);
         prev = p;
         p = p->next;
      }
   }
   fprintf(stdout, "DONE BASIC KERNEL TESTS:\n\n");
   return(bp);
}

ATL_r1node_t *ChooseKernelBlocking
(ATL_r1node_t *L1,      /* L1 blocked kernel, scope imf=0 */
 ATL_r1node_t *L2,      /* L2 blocked kernel, scope imf=1 */
 ATL_r1node_t *NOB      /* no-blocking kernel, scope imf=2 */
)
/*
 * This routine compares 3 different blocking strategy using 1-3 kernels
 * (i.e., they may all be the same kernel).
 * It is possible that the data we use may stay in a very large L3 cache,
 * so only accept no-blocking if it is significantly faster than doing
 * blocking.  L1-blocking will tend to minimize the number of kernels required,
 * so stress it very slightly more than L2 blocking.
 * RETURNS: cloned node of best blocking/kernel
 */
{
   ATL_r1node_t *best;
   double mf1, mf2, mf3;

   mf1 = L1->mflop[0] * 1.05;  /* give small adv to safest option, L1 blk */
   mf2 = L2->mflop[1] * 1.03;  /* give small adv to blocking over not */
   mf3 = NOB->mflop[2];        /* no blocking loss may vary by size, penalize */
   if (mf1 > mf2 && mf1 > mf3)
      best = L1;
   else if (mf2 > mf1 && mf2 > mf3)
      best = L2;
   else
      best = NOB;
   return(best);
}

double ExhCESrch
/*
 * RETURNS: best mflop found
 */
(
   ATL_r1node_t *r1p,           /* kernel to search with */
   char pre,                    /* type/precision prefix */
   int M, int N, int lda,
   int stride,                  /* stride to search with, real pL = pL*stride */
   int pLL,                     /* lower percL1 (mul by stride for real val) */
   int pLH,                     /* higher percL1 */
   double mfL,                  /* mflops achieved by lower */
   double mfH,                  /* mflops achieved by higher */
   int *pLB                     /* the best percL1 found */
)
{
   int plm, plb;
   double mf;

   plm = (pLH-pLL)>>1;
   if (plm < 1)
   {
      if (mfL < mfH)
      {
         *pLB = pLH;
         return(mfH);
      }
      *pLB = pLL;
      return(mfL);
   }
   plm += pLL;
   mf = TimeR2Kernel(0, 0, r1p, pre, M, N, lda, plm*stride, 0, FLUSHKB);
   fprintf(stdout, "%6d  %6d  %6d  %6d  %9.2f\n", M, N, lda, plm*stride, mf);
   mfL = ExhCESrch(r1p, pre, M, N, lda, stride, pLL, plm, mfL, mf, pLB);
   mfH = ExhCESrch(r1p, pre, M, N, lda, stride, plm, pLH, mf, mfH, &plb);
   if (mfH > mfL)
   {
      mfL = mfH;
      *pLB = plb;
   }
   return(mfL);
}

void ExhaustiveCESrch
/*
 * Performs an exhaustive search on entire range using recursive halving
 * And modifies CE to be best % of L1 size found.
 */
(
   ATL_r1node_t *r1p,           /* kernel to search with */
   int imf,                     /* set r1p->mflop[imf] to best perf */
   char pre,                    /* type/precision prefix */
   int M, int N, int lda,       /* prob size to tune with */
   int stride,                  /* stride to search with, real pL = pL*stride */
   int pLL,                     /* lower percL1 (mul by stride for real val) */
   int pLH                      /* higher percL1 */
)
{
   double mfH, mfL;
   int percL1;
   fprintf(stdout, "\nCACHE TUNING FOR %d:'%s', MU=%d NU=%d\n",
           r1p->ID, r1p->rout, r1p->XU, r1p->YU);
   fprintf(stdout, "------------------------------------------\n");
   fprintf(stdout, "     M       N     lda  percL1       mflop\n");
   fprintf(stdout, "======  ======  ======  ======  ==========\n");

   mfL = TimeR2Kernel(0, 0, r1p, pre, M, N, lda, pLL*stride, 0, FLUSHKB);
   fprintf(stdout, "%6d  %6d  %6d  %6d  %9.2f\n", M, N, lda, pLL*stride, mfL);
   mfH = TimeR2Kernel(0, 0, r1p, pre, M, N, lda, pLH*stride, 0, FLUSHKB);
   fprintf(stdout, "%6d  %6d  %6d  %6d  %9.2f\n", M, N, lda, pLH*stride, mfH);
   mfL = ExhCESrch(r1p, pre, M, N, lda, stride, pLL, pLH, mfL, mfH, &percL1);
   fprintf(stdout, "\nBEST CASE %d percent of L1, MFLOP=%.2f\n\n",
           percL1*stride, mfL);
   r1p->mflop[imf] = mfL;
   r1p->CacheElts = percL1*stride;
}

static int GetMaxID(ATL_r1node_t *r1b)
{
   ATL_r1node_t *r1p;
   int maxID=0;

   for (r1p=r1b; r1p; r1p = r1p->next)
      maxID = Mmax(maxID, r1p->ID);

   return(maxID);
}
#ifdef ATL_SSE3
void FillInGenNode
(
   ATL_r1node_t *r1p,   /* data structure to fill in */
   int nmu,             /* unrolling on mu unrolled inner (X) loop */
   int mu,              /* register blk to apply to X */
                        /* total X unrolling is nmu*mu! */
   int nu,              /* unroll&jam on outer (Y) loop */
   int evenlda,         /* assume X&Y have same (mis)align, lda is even */
   int allalign16,      /* assume X,Y,A all aligned to 16 bytes */
   int aptrs            /* use ptrs rather than lda for column indexing */
)
{
   char ln[2048];
}
ATL_r1node_t *SrchSSEGen
(
   char pre,  /* precision prefix indicating type */
   int maxID  /* IDs > maxID are safe to use in generation */
)
/*
 * This search finds the best kernel provided by the ATLAS SSE GER generator
 * RETURNS: a list of kernels to be added to the multiple implementation srch
 */
{
   return(NULL);
}
#endif

ATL_r1node_t *SortRestricted
/*
 * Sorts queue of kernels into a queue of unrestricted kernels (can be
 * always be used) and restricted (only used under certain conditions,
 * such as lda*size a multiple 16).  Destroys R1B in process.
 */
(
   ATL_r1node_t *R1B,   /* original queue containing restricted & unrest */
   ATL_r1node_t **R1R   /* queue of only restricted kernels */
)
{
   ATL_r1node_t *r1B=NULL, *r1R=NULL, *r1b=NULL, *r1r=NULL, *r1p;

   for (r1p=R1B; r1p; r1p = r1p->next)
   {
      if (r1p->ldamul > R1flag2size(r1p->flag) ||
          r1p->alignA > R1flag2size(r1p->flag) ||
          r1p->minY >= 4 || r1p->minX >= 16
         )
      {
         if (r1R)
            r1r->next = r1p;
         else
            r1R = r1p;
         r1r = r1p;
      }
      else
      {
         if (r1B)
            r1b->next = r1p;
         else
            r1B = r1p;
         r1b = r1p;
      }
   }
   if (r1R)
      r1r->next = NULL;
   assert(r1B);
   r1b->next = NULL;
   *R1R = r1R;
   return(r1B);
}

ATL_r1node_t *EliminateRepeatedRestrictions
(
   int imf,             /* entry in mflop array to read for performance */
   ATL_r1node_t *R1R    /* queue of already-timed restricted kernels */
)
/*
 * This routine sorts R1R by performance, and then eliminates all slower
 * kernels that handle the same restricted cases.
 *
 * IMPORTANT: what constitutes valid restrictions will change with
 * architectures, so we need to keep this file up-to-date.
 *
 * Things that seem like restrictions but we don't care about:
 *    alignX & Y : handled by copy in ATL_ger
 *    R1F_ALIGNX2A: handled by copy in ATL-ger
 *    R1F_ALLALIGNXY : not presently supported, but would be handled by ATL_ger
 *    R1F_FYU : handled by main loop in ATL_ger
 *
 * Valid restrictions as of 02/14/10 are:
 *    alignA: asserts A must be aligned to alignA%ATL_cachelen boundary
 *            alignA in bytes
 *    ldamul: asserts lda*size must a multiple of ldamul (which is in bytes)
 *    minY: only restriction if >= 4 (else handled by ATL_ger)
 *    minX: only restriction if >= 16 (else handled by ATL_ger)
 */
{
   ATL_r1node_t *p, *p2, *prev, *curr;

   if (!R1R)
      return(NULL);
   if (!R1R->next)
      return(R1R);
   R1R = ATL_SortR1NodesByMflop(imf, R1R);  /* sort from fastest to slowest */
   for (p=R1R; p && p->next; p = p->next)
   {
      prev = p;
      for (p2=p->next; p2; p2 = p2->next)
      {
         if (p2->alignA == p->alignA && p2->ldamul == p->ldamul)
         {
            if (p->minY < 4 && p->minX < 16)  /* quit if that's all */
               break;                         /* the restrs best kernel had */
            if (p->minY == p2->minY && p->minX == p2->minX)
               break;
         }
         prev = p2;
      }
      if (p2)  /* got a repitition */
         prev->next = KillR1Node(p2);
   }
   return(R1R);
}

ATL_r1node_t *DecimateAndRankRestricted
(
   int imf,             /* entry in mflop array to read for performance */
   ATL_r1node_t *urp,   /* unrestricted kernel to be used */
   ATL_r1node_t *R1R    /* queue of already-timed restricted kernels */
)
/*
 * Takes a timed queue of restricted kernels (with relativant results in
 * (rp->mflop[imf]), and does the following ranking:
 * (1) Any kernel not noticably outperforming urp is removed,
 * (2) Any kernels with the same restrictions are compared, and only
 *     the fastest one is retained
 * (3) Surviving queue are ranked in performance between 1 and NR,
 *     with 1 being the slowest, and NR being the fastest.  This rank
 *     is stored in rp->rankR.  When code is generated, will select kernel
 *     to use by asking if restrictions are satisfied in high-to-low order.
 * (4) RETURNS: queue of surviving kernels, sorted from fastest to slowest
 */
{
   ATL_r1node_t *rp, *prev;
/*
 * Give 4% advantage to unrestricted; not worth extra instruction load if
 * we can't get the improvement above 4% (often less than clock resolution)
 */
   const double mfur = urp->mflop[imf] * 1.04;
   int i;

/*
 * Get rid of any kernel not noticably faster than unrestricted kernel
 */
   while (R1R && R1R->mflop[imf] < mfur)
      R1R = KillR1Node(R1R);
   if (!R1R)
      return(NULL);
   for (prev=R1R,rp=R1R->next; rp; rp = rp->next)
   {
      if (rp->mflop[imf] < mfur)
      {
         prev->next = KillR1Node(rp);
         rp = prev;
      }
      else
         prev = rp;
   }
   R1R = EliminateRepeatedRestrictions(imf, R1R);
   for (i=ATL_CountNumberOfR1Nodes(R1R), rp=R1R; i > 0; rp = rp->next, i--)
      rp->rankR = i;
   return(R1R);
}

void PrintUsage(char *name, int ierr, char *flag)
{
   if (ierr > 0)
      fprintf(stderr, "Bad argument #%d: '%s'\n",
              ierr, flag ? flag : "Not enough arguments");
   else if (ierr < 0)
      fprintf(stderr, "ERROR: %s\n", flag);
   fprintf(stderr,"USAGE: %s [flags]:\n", name);
   fprintf(stderr, "   -p [s,d,c,z]: set precision prefix \n");
   fprintf(stderr,
           "   -e <fname> : external search filename (& make target)\n");
   fprintf(stderr,
           "   -o <fname> : output filename [<pre>R2K.sum]\n");
   fprintf(stderr,
           "   -i <fnam> : kernel index file [R2CASES/<pre>cases.idx]\n");
   fprintf(stderr,
           "   -C <kbytes> : default cache flush size to use\n");
   fprintf(stderr, "   -# <reps> : number of repetitions for each timing\n");
   fprintf(stderr,
      "   -#c <reps> : number of repetitions for each cache block timing\n");
   fprintf(stderr, "   -T <#> : Bitfield controlling detailed timings:\n");
   fprintf(stderr, "      0: if bitfield 0, do normal pruning search\n");
   fprintf(stderr, "      1: perform tests before timing kernels\n");
   fprintf(stderr, "      2: Time unblocked out-of-cache (OOC) performance\n");
   fprintf(stderr, "      4: Time L2-blocked OOC performance\n");
   fprintf(stderr, "      8: Time L1-blocked OOC performance\n");
   fprintf(stderr, "     16: Do Exhaustive search on OOC blocking factors\n");
   fprintf(stderr, "     32: Time in-L2 performance\n");
   fprintf(stderr, "     64: Time in-L1 performance\n");
   fprintf(stderr, "   -v <#> : verbosity level [0]\n");
   fprintf(stderr, "   -2 <kbytes> : size of L2 cache to assume\n");
   exit(ierr ? ierr : -1);
}

ATL_r1node_t *GetFlags(int nargs, char **args, char *pre, int *ESRCH,
                       char **fin, char **fout)
{
   ATL_r1node_t *eb=NULL, *ep, *en;
   int i;
   char ch;
   char ln[512];

   NREPCT = NREP = 3;
   FLUSHKB = 8192;
   VERB = 1;
   *ESRCH = 0;
   *fin = NULL;
   *fout = NULL;
   *pre = 'd';

   for (i=1; i < nargs; i++)
   {
      if (args[i][0] != '-')
         PrintUsage(args[0], i, args[i]);
      switch(args[i][1])
      {
      case 'o': /* -o <fname> : output filename [<pre>R2K.sum] */
         if (++i >= nargs)
            PrintUsage(args[0], i, NULL);
         *fout = args[i];
         break;
      case 'i': /* -i <fnam> : kernel index file [R2CASES/<pre>cases.idx] */
         if (++i >= nargs)
            PrintUsage(args[0], i, NULL);
         *fin = args[i];
         break;
      case 'v': /* -v <verbosity level> */
         if (++i >= nargs)
            PrintUsage(args[0], i, NULL);
         VERB = atoi(args[i]);
         break;
      case 'C': /* -C <kbytes> : default cache flush size to use */
         if (++i >= nargs)
            PrintUsage(args[0], i, NULL);
         FLUSHKB = atoi(args[i]);
         break;
      case '2': /* -C <kbytes> : L2 cache size to assume */
         if (++i >= nargs)
            PrintUsage(args[0], i, NULL);
         L2KB = atoi(args[i]);
         break;
      case 'T': /* -T # : do/don't do exhaustive timing */
         if (++i >= nargs)
            PrintUsage(args[0], i, NULL);
         *ESRCH = atoi(args[i]);
         break;
      case '#': /* -# <reps>/ -#c : number of repetitions for timing */
         if (++i >= nargs)
            PrintUsage(args[0], i, NULL);
         if (args[i-1][2] == 'c')
            NREPCT = atoi(args[i]);
         else
            NREP = atoi(args[i]);
         break;
      case 'e':  /* external file we should read in */
         if (++i >= nargs)
            PrintUsage(args[0], i, NULL);
         sprintf(ln, "make %s", args[i]);
         if (system(ln) != 0)
            fprintf(stderr, "\nEXTERNAL SEARCH MAKE '%s' FAILED!\n\n", ln);
         else
         {
            en = ReadR1File(ln+5);
            if (en)              /* We got new results */
            {
               if (!eb)         /* if this first, becomes base pointer */
                  eb = en;
               else             /* else, add it to queue after earlier */
               {
                  for (ep=eb; ep->next; ep = ep->next);
                  ep->next = en;
               }
            }
            else
               fprintf(stderr, "\nEMPTY EXTERNAL SEARCH '%s'!!\n\n", ln);
         }
         break;
      case 'p':  /* -p <pre> */
         if (++i >= nargs)
            PrintUsage(args[0], i, NULL);

         ch = tolower(args[i][0]);
         assert(ch == 's' || ch == 'd' || ch == 'c' || ch == 'z');
         *pre = ch;
         break;
      default:
         PrintUsage(args[0], i, args[i]);
      }
   }
/*
 * Get filenames in freeable string
 */
   if (!(*fout))
   {
      if (*ESRCH)
         sprintf(ln, "res/%cR2Ktimes.sum", *pre);
      else
         sprintf(ln, "res/%cR2K.sum", *pre);
      *fout = DupString(ln);
   }
   else
      *fout = DupString(*fout);
   if (!(*fin))
   {
      if (*ESRCH)
         sprintf(ln, "R2CASES/%cr2time.idx", *pre);
      else
         sprintf(ln, "R2CASES/%cr2cases.idx", *pre);
      *fin = DupString(ln);
   }
   else
      *fin = DupString(*fin);
/*
 * Don't trust external search's own timings & settings: we will test ourselves
 */
   for (ep=eb; ep; ep = ep->next)
   {
      for (i=0; i < 8; i++)
         ep->mflop[0] = 0.0;
      if (ep->kname)
         free(ep->kname);
      ep->kname = NULL;
      ep->rankR = 0;
   }
   return(eb);
}

void WriteMflopExp(FILE *fpout)
{
   fprintf(fpout, "#\n#MFLOP array has following meaning by index:\n");
   fprintf(fpout, "#   0 : Out-of-cache, L1 blocked\n");
   fprintf(fpout, "#   1 : Out-of-cache, L2 blocked\n");
   fprintf(fpout, "#   2 : Out-of-cache, no blocking\n");
   fprintf(fpout, "#   3 : Problem preloaded to L2, no blocking\n");
   fprintf(fpout, "#   4 : Problem preloaded to L1, no blocking\n#\n");
}

void WriteR1SummFile(char *fnam, ATL_r1node_t *r1b)
{
   char ln[256];
   FILE *fpout;
   ATL_r1node_t *bases[4], *r1p;
   char *exp[4] = {"out-of-cache operands",
                   "operands preloaded to the L2 cache",
                   "operands preloaded to the L1 cache",
                   "operands out-of-cache, but blocked"
                  };
   int i;

   ATL_R1SplitContexts(r1b, bases, bases+1, bases+2, bases+3);
   fpout = fopen(fnam, "w");
   assert(fpout);
   WriteMflopExp(fpout);
   fprintf(fpout, "#\n#Each kernel context has multiple kernels:\n");
   fprintf(fpout,
           "#   All kernels with rankR > 0 have restrictions, and are only\n");
   fprintf(fpout,
           "#   used when certain conditions are met.  The last kernel in\n");
   fprintf(fpout,
"#   the context series has rankR=0, and can be called for any valid input\n");
   fprintf(fpout, "#\n");
   for (i=0; i < 4; i++)
   {
      fprintf(fpout,
"# ------------------------------------------------------------------------\n");

      fprintf(fpout,
         "# Following %d GER2 kernels are optimized for %s\n",
              bases[i]->rankR+1, exp[i]);
      if (i == 3)
         fprintf(fpout,
              "# for L1 reuse.  These are used in low-rank update\n");
      fprintf(fpout,
"# ------------------------------------------------------------------------\n");
      for (r1p = bases[i]; r1p; r1p = r1p->next)
         PrintR1Line(fpout, r1p);
   }
   fclose(fpout);
/*
 * Restore queue in case it used by caller
 */
   r1b = ATL_R1LinkContexts(bases[0],bases[1],bases[2],bases[3]);
}

void NameAllKernels(char pre, ATL_r1node_t *r1b)
/*
 * Gives all kernels unique names.  For now, just name them gerk_<ID>
 */
{
   ATL_r1node_t *p;
   char ln[64];

   for (p=r1b; p; p = p->next)
   {
      sprintf(ln, "ATL_%cger2k__%d", pre, p->ID);
      p->kname = DupString(ln);
   }
}

static int Mylcm(const int M, const int N)
/*
 * Returns least common multiple (LCM) of two positive integers M & N by
 * computing greatest common divisor (GCD) and using the property that
 * M*N = GCD*LCM.
 */
{
   register int tmp, max, min, gcd=0;

   if (M != N)
   {
      if (M > N) { max = M; min = N; }
      else { max = N; min = M; }
      if (min > 0)  /* undefined for negative numbers */
      {
         do  /* while (min) */
         {
            if ( !(min & 1) ) /* min is even */
            {
               if ( !(max & 1) ) /* max is also even */
               {
                  do
                  {
                     min >>= 1;
                     max >>= 1;
                     gcd++;
                     if (min & 1) goto MinIsOdd;
                  }
                  while ( !(max & 1) );
               }
               do min >>=1 ; while ( !(min & 1) );
            }
/*
 *          Once min is odd, halve max until it too is odd.  Then, use
 *          property that gcd(max, min) = gcd(max, (max-min)/2)
 *          for odd max & min
 */
MinIsOdd:
            if (min != 1)
            {
               do  /* while (max >= min */
               {
                  max -= (max & 1) ? min : 0;
                  max >>= 1;
               }
               while (max >= min);
            }
            else return( (M*N) / (1<<gcd) );
            tmp = max;
            max = min;
            min = tmp;
         }
         while(tmp);
      }
      return( (M*N) / (max<<gcd) );
   }
   else return(M);
}

ATL_r1node_t *FindSmallestNU(ATL_r1node_t *r1b)
/*
 * RETURNS: ptr to node wt smallest NU setting
 */
{
   ATL_r1node_t *r1min, *r1p;
   int numin;

   if (!r1b)
      return(NULL);
   numin = r1b->YU;
   r1min = r1b;
   for (r1p=r1b; r1p; r1p = r1p->next)
   {
      if (r1p->YU < numin)
      {
         numin = r1p->YU;
         r1min = r1p;
      }
   }
   return(r1min);
}

ATL_r1node_t *SortQByNU(ATL_r1node_t *r1b)
/*
 * Sorts queue on NU (greatest-to-least)
 * RETURNS: pointer to new Q sorted from greatest to least on NU (YU)
 */
{
   ATL_r1node_t *r1s=NULL, *r1p, *r1m;
   while (r1b)
   {
      r1m = FindSmallestNU(r1b);
      if (r1m == r1b)
         r1b = r1b->next;
      else
      {
         for (r1p=r1b; r1p->next != r1m; r1p = r1p->next);
         r1p->next = r1p->next->next;
      }
      r1m->next = r1s;
      r1s = r1m;
   }
   return(r1s);
}

void ExhaustiveTime(char pre, char *fin, char *fout, int lvl)
/*
 * This routine attempts to time every kernel for all contexts, with the
 * out-of-cache blocking being individually tuned for each routine.  It is
 * used to compare differing implementations head-to-head, and does no
 * kernel selection, just the timing.
 * All timings use all operands aligned to ATL_Cachelen, incX=incY=1.
 */
{
   ATL_r1node_t *r1b, *r1p, *r1OOC, *r1L1, *r1L2, *r1s;
   ATL_r1node_t *r1L1b=NULL, *r1L2b=NULL;
   double mfB;
   char *blkstr;
   FILE *fpout;
   size_t Nooc=8000/16, Mooc=8000*16;  /* rough dims of out-of-cache problem */
   int Nl2, Ml2, Nl1, Ml1, lcmM, lcmN, minNU, maxNU, n;
   int L1ce, L2ce, fflops, L2CacheElts, L1CacheElts, i;

   r1b = ReadR1File(fin);  /* get kernels to be timed */
   assert(r1b);
   r1b = DelRepeatedR1Kernels(r1b);
   r1b = DelBadArchR1Kernels(r1b);
   if (lvl & 1)
      r1b = DelBadTestKernels(pre, r1b);
   assert(r1b);
   fprintf(stdout, "\nSurviving cases:\n");
   WriteR1File("stdout", r1b);
/*
 * Find M/N unrollings that are multiples of all kernel's native blocking
 * Force lcmM to be at least 32 bytes long to avoid alignment issues
 */
   if (pre == 's')
      lcmM = 8;
   else if (pre == 'd' || pre == 'c')
      lcmM = 4;
   else
      lcmM = 2;
   lcmN = 1;
   maxNU = minNU = r1b->YU;
   for (r1p=r1b; r1p; r1p = r1p->next)
   {
      lcmM = Mylcm(lcmM, r1p->XU);
      lcmN = Mylcm(lcmN, r1p->YU);
      if (minNU > r1p->YU)
         minNU = r1p->YU;
      if (maxNU < r1p->YU)
         maxNU = r1p->YU;
   }
/*
 * Set out-of-cache dimensions; first compute minimum acceptable values
 */
   L2CacheElts = (L2KB) ? (L2KB*1024)/pre2size(pre) : 4*L1CacheElts;
   Mooc = (1.1*L2CacheElts)/(minNU+1);  /* OOC working set for everyone */
   Mooc = ((Mooc+lcmM-1)/lcmM)*lcmM;
   Nooc = ((8*maxNU+lcmN-1)/lcmN)*lcmN;
   if (Nooc < 100)
      Nooc = ((100+lcmN-1)/lcmN)*lcmN;
/*
 * Adjust M upward until matrix at least as big as 8000x8000
 */
   if (Mooc * Nooc < 8000*8000)
   {
      Mooc = (8000*8000)/Nooc;
      Mooc = ((Mooc+lcmM-1)/lcmM)*lcmM;
   }
/*
 * If matrix bigger than 488MB, may cause swapping to time it, so
 * reduce Nooc, at the cost of having to floor it with NU
 */
   if (Mooc*Nooc*pre2size(pre) > 8000*8000*sizeof(double))
   {
      Nooc = (8000*8000*sizeof(double))/pre2size(pre);
      Nooc = (Nooc)/Mooc;
      Nooc = ((Nooc+maxNU-1)/maxNU)*maxNU;
      if (Nooc < maxNU)
         Nooc = maxNU;
   }


/*
 * Force in-cache problems to do 3 mflops for timing
 */
   fflops = 3;
/*
 * Set in-L2 cache problem dimensions, estimate use of 80% of cache
 */
   Nl2 = lcmN;
   Ml2 = 0.8*L2CacheElts;
   Ml2 /= Nl2;
   Ml2 = (Ml2/lcmM)*lcmM;
/*
 * If M dimension strongly constrained, rescale N dim and accept slightly
 * differing problem sizes.  Need Q sorted from greatest to least, so that
 * we ensure the common N can use the floor operation with YU in order
 * to vary the N dim to work with all YU
 */
   if (Ml2 < 480)
   {
      r1s = CloneR1Queue(r1b);  /* get dup Q */
      r1s = SortQByNU(r1s);     /* sort from greatest to least */
      Ml2 = 0.8*L2CacheElts;
      i = Nl2 = r1s->YU;
      for (r1p=r1s->next; r1p; r1p = r1p->next)
      {
         i = Mylcm(i, r1p->YU);
         if (i*480 > Ml2) break;
         Nl2 = i;
      }
      KillAllR1Nodes(r1s);
      Ml2 /= Nl2;
      Ml2 = (Ml2/lcmM)*lcmM;
   }
/*
 * Set in-L1 cache problem dimensions, estimate use of 90% of cache
 */
   L1CacheElts = (GetL1CacheSize()*1024) / pre2size(pre);
   Nl1 = lcmN;
   Ml1 = 0.9*L1CacheElts;
   Ml1 /= Nl1;
   Ml1 = (Ml1/lcmM)*lcmM;
   if (Ml1 < 120)
   {
      r1s = CloneR1Queue(r1b);  /* get dup Q */
      r1s = SortQByNU(r1s);     /* sort from greatest to least */
      Ml1 = 0.9*L2CacheElts;
      i = Nl1 = r1s->YU;
      for (r1p=r1s->next; r1p; r1p = r1p->next)
      {
         i = Mylcm(i, r1p->YU);
         if (i*120 > Ml1) break;
         Nl1 = i;
      }
      KillAllR1Nodes(r1s);
      Ml1 /= Nl1;
      Ml1 = (Ml1/lcmM)*lcmM;
   }
   printf("Mooc=%d, Nooc=%d;  ML2=%d ,NL2=%d;  ML1=%d NL1=%d\n",
          Mooc, Nooc, Ml2, Nl2, Ml1, Nl1);
   if (lvl & 2)
   {
/*
 *    Find kernels unblocked performance for out-of-cache timings
 */
      printf("\nTIMING OUT-OF-CACHE CASES WITH NO BLOCKING, M=%d, N=%d\n",
             Mooc, Nooc);
      for (r1p=r1b; r1p; r1p = r1p->next)
      {
         r1p->CacheElts = 0;
         r1p->mflop[2] = TimeR2Kernel(VERB, 0, r1p, pre, Mooc,
                                      (Nooc/r1p->YU)*r1p->YU, Mooc, 0,
                                      fflops, 0);
      }
   }
/*
 * Handle blocking tuning differently if we need exhaustive search
 */
   if (lvl & 16)
   {
      if (lvl & 8) /* time L1-blocked OOC? */
      {
         r1L1b = CloneR1Queue(r1b);  /* get dup Q */
         for (r1p=r1L1b, r1s=r1b; r1p; r1p = r1p->next, r1s=r1s->next)
         {
            n = (Nooc/r1p->YU)*r1p->YU;
            ExhaustiveCESrch(r1p, 0, pre, Mooc, n, Mooc, 2, 25, 50);
            r1p->CacheElts = L1CacheElts*0.01*r1p->CacheElts;
            if (r1p->mflop[0] >= r1s->mflop[2])
               r1s->CacheElts = r1p->CacheElts;
            r1s->mflop[0] = r1p->mflop[0];
         }
      }
      if (lvl & 4)  /* time L2-blocked OOC? */
      {
         r1L2b = CloneR1Queue(r1b);  /* get dup Q */
         for (r1p=r1L2b, r1s=r1b; r1p; r1p = r1p->next, r1s = r1s->next)
         {
            n = (Nooc/r1p->YU)*r1p->YU;
            r1p->CacheElts = L2CacheElts;
            i = 2*L2CacheElts / (L1CacheElts);
            i = Mmax(i, 16);
            ExhaustiveCESrch(r1p, 1, pre, Mooc, n, Mooc, 50, 3, i);
            r1p->CacheElts = L1CacheElts*0.01*r1p->CacheElts;
            if (r1p->mflop[1] >= r1s->mflop[2] && r1p->mflop[1] > r1s->mflop[0])
               r1s->CacheElts = r1p->CacheElts;
            r1s->mflop[1] = r1p->mflop[1];
         }
      }
      if (lvl & 2)  /* Timed no-blocking results */
      {
         printf("\nNO BLOCKING RESULTS:\n");
         printf("--------------------\n");
         for (r1p=r1b; r1p; r1p = r1p->next)
            printf("   %d:%s, M=%d, N=%d, CacheElts=%d, gets =%.2f MFLOPS\n",
                   r1p->ID, r1p->rout, Mooc, (Nooc/r1p->YU)*r1p->YU,
                   r1p->CacheElts, r1p->mflop[2]);
      }
      if (lvl & 8) /* timed L1-blocked OOC? */
      {
         printf("\nL1 BLOCKING RESULTS:\n");
         printf("--------------------\n");
         for (r1p=r1L1b; r1p; r1p = r1p->next)
            printf("   %d:%s, M=%d, N=%d, CacheElts=%d, gets =%.2f MFLOPS\n",
                   r1p->ID, r1p->rout, Mooc, (Nooc/r1p->YU)*r1p->YU,
                   r1p->CacheElts, r1p->mflop[0]);
      }
      if (lvl & 4) /* timed L2-blocked OOC? */
      {
         printf("\nL2 BLOCKING RESULTS:\n");
         printf("--------------------\n");
         for (r1p=r1L2b; r1p; r1p = r1p->next)
            printf("   %d:%s, M=%d, N=%d, CacheElts=%d, gets =%.2f MFLOPS\n",
                   r1p->ID, r1p->rout, Mooc, (Nooc/r1p->YU)*r1p->YU,
                   r1p->CacheElts, r1p->mflop[1]);
      }
   }
   else  /* Just do normal 75/90% searches */
   {
      if (lvl & 8) /* time L1-blocked OOC? */
      {
/*
 *       Find kernels L1-blocked & L2-blocked performance using 85% of cache
 */
         printf("\nTIMING OUT-OF-CACHE CASES WITH L1 BLOCKING, M=%d, N=%d\n",
                Mooc, Nooc);
         for (r1p=r1b; r1p; r1p = r1p->next)
         {
            r1p->CacheElts = 0.92 * L1CacheElts;
            r1p->mflop[0] = TimeR2Kernel(VERB, 0, r1p, pre, Mooc,
                                         (Nooc/r1p->YU)*r1p->YU, Mooc, 85,
                                         fflops, 0);
         }
      }
      if (lvl & 4) /* time L2-blocked OOC? */
      {
         printf("\nTIMING OUT-OF-CACHE CASES WITH L2 BLOCKING, M=%d, N=%d\n",
                Mooc, Nooc);
         for (r1p=r1b; r1p; r1p = r1p->next)
         {
            r1p->CacheElts = 0.75 * L2CacheElts;
            i = (100.0*r1p->CacheElts)/(1.0*L1CacheElts);
            r1p->mflop[1] = TimeR2Kernel(VERB, 0, r1p, pre, Mooc,
                                         (Nooc/r1p->YU)*r1p->YU, Mooc, i,
                                         fflops, 0);
         }
      }
      if (lvl&12)  /* did both L1 & L2, choose best one */
      {
         for (r1p=r1b; r1p; r1p = r1p->next)
            r1p->CacheElts = (r1p->mflop[0] >= r1p->mflop[1]) ?
                             0.92*L1CacheElts : 0.75*L2CacheElts;
      }
   }  /* end if on exhaustive or estimate blocking search */

   if (lvl & 32) /* time in-L2 problems? */
   {
      printf("\nTIMING IN-L2 PROBLEMS, M=%d, N=%d\n", Ml2, Nl2);
      for (r1p=r1b; r1p; r1p = r1p->next)
         r1p->mflop[3] = TimeR2Kernel(VERB, 0, r1p, pre, Ml2,
                                      (Nl2/r1p->YU)*r1p->YU, Ml2, 0, fflops, 0);
   }
   if (lvl & 64) /* time in-L1 performance? */
   {
      printf("\nTIMING IN-L1 PROBLEMS, M=%d, N=%d\n", Ml1, Nl1);
      for (r1p=r1b; r1p; r1p = r1p->next)
         r1p->mflop[4] = TimeR2Kernel(VERB, 0, r1p, pre, Ml1,
                                      (Nl1/r1p->YU)*r1p->YU, Ml1, 0, fflops, 0);
   }
/*
 * Find best kernel for out-of-cache, and the blocking it used
 */
   r1OOC = NULL;
   mfB = 0.0;

   if (lvl & 2)  /* did we time unblocked OOC? */
   {
      r1OOC = FindFastestKernel(pre, r1b, 2, 1);
      r1OOC->CacheElts = 0;
      blkstr = "no blocking";
      mfB = r1OOC->mflop[2];
   }
   if (lvl&16)  /* did exhaustive cache blocking search */
   {
      if (lvl & 4) /* did we time L2-blocked OOC? */
      {
         r1p = FindFastestKernel(pre, r1L2b, 1, 1);
         if (!r1OOC || r1p->mflop[1] > mfB)
         {
            r1OOC = CloneR1Node(r1p);
            r1OOC->next = NULL;
            blkstr = "L2 blocking";
            mfB = r1p->mflop[1];
         }
         KillAllR1Nodes(r1L2b);
      }

      if (lvl & 8) /* did we time L1-blocked OOC? */
      {
         r1p = FindFastestKernel(pre, r1L1b, 0, 1);
         if (!r1OOC || r1p->mflop[0] > mfB)
         {
            r1OOC = CloneR1Node(r1p);
            r1OOC->next = NULL;
            blkstr = "L1 blocking";
            mfB = r1p->mflop[0];
         }
         KillAllR1Nodes(r1L1b);
      }
   }
   else  /* did simple % estimate cacheedge search */
   {
      if (lvl & 4) /* did L2-blocked OOC */
      {
         r1p = FindFastestKernel(pre, r1b, 1, 1);
         if (!r1OOC || r1p->mflop[1] > mfB)
         {
            mfB = r1p->mflop[1];
            r1OOC = r1p;
            blkstr = "L2 blocking";
         }
      }
      if (lvl & 8) /* did L1-blocked OOC */
      {
         r1p = FindFastestKernel(pre, r1b, 0, 1);
         if (!r1OOC || r1p->mflop[0] > mfB)
         {
            mfB = r1p->mflop[0];
            r1OOC = r1p;
            blkstr = "L1 blocking";
         }
      }
   }    /* end if on type of cacheedge search */
/*
 * Find best in-L2 and in-L1 kernels
 */
   r1L2 = FindFastestKernel(pre, r1b, 3, 1);
   r1L1 = FindFastestKernel(pre, r1b, 4, 1);

   fpout = fopen(fout, "w");
   assert(fpout);
   WriteMflopExp(fpout);
   fprintf(fpout,
           "#\n# Mooc=%d, Nooc=%d, Ml2=%d, Nl2=%d, Ml1=%d Nl1=%d\n#\n#\n",
           Mooc, Nooc, Ml2, Nl2, Ml1, Nl1);
   fprintf(fpout, "# ==========================\n");
   fprintf(fpout, "# Performance of all kernels\n");
   fprintf(fpout, "# ==========================\n");
   PrintR1Nodes(fpout, r1b);
   fprintf(fpout, "################################### -- CUT HERE -- ############################\n");
   fprintf(fpout, "###############################################################################\n#\n");

   if (lvl & 64)
   {
      fprintf(fpout, "# ----------------------------------------\n");
      fprintf(fpout, "# Following kernel is best in-L1 performer\n");
      fprintf(fpout, "# ----------------------------------------\n");
      PrintR1Line(fpout, r1L1);
   }
   if (lvl & 32)
   {
      fprintf(fpout, "# ----------------------------------------\n");
      fprintf(fpout, "# Following kernel is best in-L2 performer\n");
      fprintf(fpout, "# ----------------------------------------\n");
      PrintR1Line(fpout, r1L2);
   }
   if (lvl & 2)
   {
      fprintf(fpout, "# ---------------------------------------------------\n");
      fprintf(fpout, "# Following kernel is best out-of-cache using %s\n",
              blkstr);
      fprintf(fpout, "# ---------------------------------------------------\n");
      PrintR1Line(fpout, r1OOC);
   }
   WriteMflopExp(fpout);
   fclose(fpout);
}

main(int nargs, char **args)
{
   ATL_r1node_t *ocb, *i2b, *i1b, *syb;
   ATL_r1node_t *r1b, *r1p, *r1r, *r1bestL2b;
   ATL_r1node_t *r1bestOC, *r1bestOCr, *r1bestL1b, *r1bestL1br;
   ATL_r1node_t *r1bestL1, *r1bestL1r, *r1bestL2, *r1bestL2r;
   FILE *fpin, *fpout;
   double mf, percL1;
   int i, L1CacheElts, CE1, CE2, maxID, bigN, ETIME;
   char *fin, *fout;
   char ln[128];
   char pre;

   r1r = GetFlags(nargs, args, &pre, &ETIME, &fin, &fout);
   if (ETIME)
   {
      ExhaustiveTime(pre, fin, fout, ETIME);
      exit(0);
   }
   if (pre == 'd' || pre == 'c')
      bigN = 2300;
   else if (pre == 's')
      bigN = 3000;
   else
      bigN = 1500;
   L1CacheElts = (GetL1CacheSize()*1024) / pre2size(pre);
   r1b = ReadR1File(fout);
   SetAllR1TypeFlags(pre, r1b);
   maxID = GetMaxID(r1b);
   if (r1b)
   {
      if (r1r)  /* don't care about external searches if we've got answer */
         KillAllR1Nodes(r1r);
      if (r1b->mflop[3] <= 0.0)
      {
/*
 *       Split queue by context/routine
 */
         ATL_R1SplitContexts(r1b, &ocb, &i2b, &i1b, &syb);
/*
 *       Retime out-of-cache kernel
 */
         TimeAllKernels(L1CacheElts, 0, pre, bigN, bigN, bigN, 0, 0, -1, ocb);
/*
 *       Retime in-L2 timings
 */
         TimeAllKernelsForContext(L1CacheElts, 3, pre, i2b);
/*
 *       Retime in-L1 timings
 */
         TimeAllKernelsForContext(L1CacheElts, 4, pre, i1b);
/*
 *       Retime L1-blocked out-of-cache timings
 */
         TimeAllKernels(L1CacheElts, 0, pre, bigN, bigN, bigN, 0, 0, -1, syb);
         r1b = ATL_R1LinkContexts(ocb, i2b, i1b, syb);
         WriteR1SummFile(fout, r1b);
      }
      PrintR1Nodes(stdout, r1b);
      KillAllR1Nodes(r1b);
      exit(0);
   }
   r1b = ReadR1File(fin);
   if (r1r)  /* add external outputs to our queue */
   {
      if (!r1b)
         r1b = r1r;
      else
      {
         for (r1p=r1b; r1b->next; r1p = r1p->next);
         r1p->next = r1r;
      }
   }
   if (!r1b)
   {
      fprintf(stderr, "\n%s: NO INPUT KERNELS, CANNOT RUN!\n\n", args[0]);
      exit(-1);
   }
   fprintf(stdout, "\nCases read in:\n");
   WriteR1File("stdout", r1b);
   r1b = DelRepeatedR1Kernels(r1b);
   r1b = DelBadArchR1Kernels(r1b);
   r1b = DelBadTestKernels(pre, r1b);
   fprintf(stdout, "\nSurviving cases:\n");
   WriteR1File("stdout", r1b);

   #ifdef ATL_SSE3   /* current generator requires SSE3 */
      r1p = SrchSSEGen(pre, maxID);
      if (r1p)
      {
         while (r1p->next)
            r1p = r1p->next;
         r1p->next = r1b;
         r1b = r1p;
      }
   #endif
   r1b = SortRestricted(r1b, &r1r);     /* sort into general & rest kernels */
/*
 * Find best general kernel for L1-blocked out-of-cache behavior; use
 * cache elements of 85% of L1 size
 */
   for (r1p=r1b; r1p; r1p = r1p->next)
      r1p->CacheElts = 0.85 * L1CacheElts;
   fprintf(stdout, "\nBEGIN L1-BLOCKED TUNING\n");
   r1p = TimeAllKernels(L1CacheElts, 0, pre, bigN, bigN, bigN, 0, 0, -1, r1b);
   r1bestL1b = CloneR1Node(r1p);
   r1bestL1b->next = NULL;
   fprintf(stdout, "DONE L1-BLOCKED TUNING, CHOSE '%s' (%.2f)\n\n",
           r1bestL1b->rout, r1bestL1b->mflop[0]);
/*
 * Find best L1 blocking % of cache, convert back to Elts
 */
   ExhaustiveCESrch(r1bestL1b, 0, pre, bigN, bigN, bigN, 2, 25, 50);
   r1bestL1b->CacheElts = L1CacheElts*0.01*r1bestL1b->CacheElts;
/*
 * Find best kernel for L2-blocked out-of-cache context; use min of 4*L1
 * and 128K as good effective L2 estimate, convert back to elts
 */
   CE1 = 4 * L1CacheElts;
   CE1 = Mmin(CE1, 128*1024/pre2size(pre));
   for (r1p=r1b; r1p; r1p = r1p->next)
      r1p->CacheElts = CE1;
   fprintf(stdout, "\nBEGIN L2-BLOCKED TUNING\n");
   r1p = TimeAllKernels(L1CacheElts, 1, pre, bigN, bigN, bigN, 0, 0, -1, r1b);
   r1bestL2b = CloneR1Node(r1p);
   r1bestL2b->next = NULL;
   fprintf(stdout, "DONE L2-BLOCKED TUNING, CHOSE '%s' (%.2f)\n\n",
           r1bestL2b->rout, r1bestL2b->mflop[1]);

   ExhaustiveCESrch(r1bestL2b, 1, pre, bigN, bigN, bigN,
                    50, 3, 16);
   r1bestL2b->CacheElts = L1CacheElts*0.01*r1bestL2b->CacheElts;

   printf("BEST L1-blocked kernel:\n");
   WriteR1File("stdout", r1bestL1b);
   printf("BEST L2-blocked kernel:\n");
   WriteR1File("stdout", r1bestL2b);
/*
 * Find best kernel for in-L2 and in-L1 usage
 */
   r1bestL1 = TimeAllKernelsForContext(L1CacheElts, 4, pre, r1b);
   r1bestL2 = TimeAllKernelsForContext(L1CacheElts, 3, pre, r1b);
   r1bestOC = TimeAllKernelsForContext(L1CacheElts, 2, pre, r1b);
   r1bestL1 = CloneR1Node(r1bestL1);
   r1bestL2 = CloneR1Node(r1bestL2);
   r1bestL2->CacheElts = r1bestL2b->CacheElts;
   r1bestOC = CloneR1Node(r1bestOC);
   r1bestOC->CacheElts = 0;
   r1bestL1->next = r1bestL2->next = r1bestOC->next = NULL;
   r1bestL1->CacheElts = r1bestL1b->CacheElts;
/*
 * Figure out what type of blocking to use for out-of-cache context
 */
   r1p = ChooseKernelBlocking(r1bestL1, r1bestL2, r1bestOC);
   if (r1p != r1bestOC)
   {
      KillR1Node(r1bestOC);
      r1bestOC = CloneR1Node(r1p);
   }
/*
 * When timing restricted kernels, we will use best CE found by unresticted,
 * and we can tune only the contexts we use: r1bestOC, r1bestL1[b], r1bestL2.
 */
   if (r1r)
   {
/*
 *    Find best restricted out-of-cache, L1blocked, in-L1 & in-L2 kernels
 */
      fprintf(stdout, "\nBEGIN RESTRICTED OUT-OF-CACHE TUNING\n");
      for (r1p=r1r; r1p; r1p = r1p->next)
         r1p->CacheElts = r1bestOC->CacheElts;
      if (r1bestOC->CacheElts > L1CacheElts)
         i = 1;
      else if (r1bestOC->CacheElts)
         i = 0;
      else
         i = 2;
      TimeAllKernels(L1CacheElts, i, pre, bigN, bigN, bigN, 0, 0, -1, r1r);
      r1p = CloneR1Queue(r1r);  /* get private copy for destruction */
      r1bestOCr = DecimateAndRankRestricted(i, r1bestOC, r1p);
      if (r1bestOCr)
      {
         fprintf(stdout, "KERNEL CHOSEN:\n");
         PrintR1Nodes(stdout, r1bestOCr);
         fprintf(stdout, "DONE RESTRICTED OUT-OF-CACHE TUNING.\n");
      }
      else
         fprintf(stdout, "NO RESTRICTED OUT-OF-CACHE KERNEL USED.\n");

      fprintf(stdout, "\nBEGIN RESTRICTED L1-BLOCKED TUNING\n");
      for (r1p=r1r; r1p; r1p = r1p->next)
         r1p->CacheElts = r1bestL1b->CacheElts;
      TimeAllKernels(L1CacheElts, 0, pre, bigN, bigN, bigN, 0, 0, -1, r1r);
      r1p = CloneR1Queue(r1r);  /* get private copy for destruction */
      r1bestL1br = DecimateAndRankRestricted(0, r1bestOC, r1p);
      if (r1bestL1br)
      {
         fprintf(stdout, "KERNEL CHOSEN:\n");
         PrintR1Nodes(stdout, r1bestL1br);
         fprintf(stdout,"DONE RESTRICTED L1-BLOCKED TUNING.\n");
      }
      else
         fprintf(stdout, "NO RESTRICTED L1-BLOCKED KERNEL USED.\n");

      fprintf(stdout, "\nBEGIN RESTRICTED in-L1 TUNING\n");
      TimeAllKernelsForContext(L1CacheElts, 4, pre, r1r);
      r1bestL1r = CloneR1Queue(r1r);  /* get private copy for destruction */
      r1bestL1r = DecimateAndRankRestricted(4, r1bestOC, r1bestL1r);
      if (r1bestL1r)
      {
         fprintf(stdout, "KERNEL CHOSEN:\n");
         PrintR1Nodes(stdout, r1bestL1r);
         fprintf(stdout, "DONE RESTRICTED in-L1 TUNING.\n");
      }
      else
         fprintf(stdout, "NO RESTRICTED L1-BLOCKED KERNEL USED.\n");

      fprintf(stdout, "\nBEGIN RESTRICTED in-L2 TUNING\n");
      for (r1p=r1r; r1p; r1p = r1p->next)
         r1p->CacheElts = r1bestL2b->CacheElts;
      TimeAllKernelsForContext(L1CacheElts, 3, pre, r1r);
      r1bestL2r = CloneR1Queue(r1r);  /* get private copy for destruction */
      r1bestL2r = DecimateAndRankRestricted(3, r1bestOC, r1bestL2r);
      if (r1bestL1r)
      {
         fprintf(stdout, "KERNEL CHOSEN:\n");
         PrintR1Nodes(stdout, r1bestL2r);
         fprintf(stdout, "DONE RESTRICTED in-L2 TUNING.\n");
      }
      else
         fprintf(stdout, "NO RESTRICTED in-L2 KERNEL USED.\n");
   }
   else
      r1bestOCr = r1bestL1r = r1bestL2r = r1bestL1br = NULL;
/*
 * Indicate unrestricted kernel by setting rankR=0, and add it to end
 * of restricted kernels to make our final fastest-to-slowest call chain
 */
   KillAllR1Nodes(r1b);
   KillAllR1Nodes(r1r);
   r1bestOC->rankR = r1bestL1b->rankR = r1bestL1->rankR = r1bestL2->rankR = 0;

   if (r1bestOCr)
   {
      r1p = ATL_LastR1Node(r1bestOCr);
      r1p->next = r1bestOC;
   }
   else
      r1bestOCr = r1bestOC;

   if (r1bestL1br)
   {
      r1p = ATL_LastR1Node(r1bestL1br);
      r1p->next = r1bestL1b;
   }
   else
      r1bestL1br = r1bestL1b;

   if (r1bestL1r)
   {
      r1p = ATL_LastR1Node(r1bestL1r);
      r1p->next = r1bestL1;
   }
   else
      r1bestL1r = r1bestL1;

   if (r1bestL2r)
   {
      r1p = ATL_LastR1Node(r1bestL2r);
      r1p->next = r1bestL2;
   }
   else
      r1bestL2r = r1bestL2;
/*
 * Join all together in combined context queue, and write out answer
 */
   r1b = ATL_R1LinkContexts(r1bestOCr, r1bestL2r, r1bestL1r, r1bestL1b);
   NameAllKernels(pre, r1b);
   WriteR1SummFile(fout, r1b);
   KillAllR1Nodes(r1b);
   return(0);
}

